/*
 **********************************************************************
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 **********************************************************************

Soekris Net45xx GPIO driver
Provides basic control (read and write) over the 8 general purpose
IO-pins of the Soekris Net4501 and the error-led (which happens to
be connected to another general purpose IO pin)
LEAF package by __PACKAGER__, __BUILDDATE__

Accessing the GPIO pins (assuming all are output):
        echo 11111111 > /proc/driver/soekris_gpio --> turns all ports on
        echo 00000000 > /proc/driver/soekris_gpio --> turns all ports off
        echo 00000100 > /proc/driver/soekris_gpio --> turns GPIO2 on,
                                                      everything else off

        cat /proc/driver/soekris_gpio   --> prints the current
                                            status of the gpio pins

        Note when GPIO are used as inputs:
        when "dangling" (not connected to anything)
        GPIO0 to GPIO3 are seen as "0"
        GPIO 4 to GPIO7 are seen as "1"

        accessing the error LED:
        echo 1 > /proc/driver/soekris_error_led --> turns the led on

        cat /proc/driver/soekris_error_led
                 --> prints the current status of the error led

Setting input or output for a pin
        echo 10000001 /proc/driver/soekris_io_settings ->
                        turns GPIO7 and GPIO0 to output, all the others to input

        cat /proc/driver/soekris_io_settings -->prints the current io-settings

Devices:
        major ? (dynamic) minor 0 :   set/read PIO0-7 (Net4501 mapping)
        major ? (dynamic) minor 254 : set/read error_led (Net4501)
        Which major number has been assigned can be determined by looking at
        /proc/devices


Known IOCTL commands:
        PPCLAIM
        PPRELEASE
        PPFCONTROL
                    These don't do anything, but don't cause an error either

        PPDATADIR
                    Data line direction for GPIO0-7: non-zero for input mode.
                    This is _not_ "generic" at all, but very much hard-wired towards
                    being able to use an HD44780 LCD on the GPIO pins (in 4-bit mode)
                    and being able to do so using generic ppdev instructions

                    So, GPIO4-GPIO7 will _always_ be set to output after this call, only
                    GPIO0-GPIO3 are changed

        PPWDATA,PPRDATA
                    Reads/writes to the gpio0-7 (minor 0)
                    or the error led (minor 254)

        GPIORDDIRECTION,GPIOWRDIRECTION
                    "new" command (not from parport) - same effect
                    as using /proc/driver/soekris_io_settings

        GPIORDDATA,GPIOWRDATA
                    "new" command (not from parport)
                    read/write GPIO0-7


Credits/References
        Linux Device Drivers, 2nd Edition, O'Reilly, ISBN 0-59600-008-1
        http://www.blurbco.com/~gork/net4501/net4501_modules-1.2.tar.gz
        linux-2.4.20/drivers/char/sc520_wdt.c
        http://www.tldp.org/LDP/lkmpg/
        Soekris Engineering net4501 User's Manual
        Elan SC520 Microcontroller User's Manual
        Elan SC520 Microcontroller Register Set Manual
        soekris-tech mailing list
                (http://lists.soekris.com/mailman/listinfo/soekris-tech)

(c) Copyright 2003   Martin Hejl <martin.hejl@guh-software.de>
                     G&H Softwareentwicklung GmbH

Changelog
1.0   2003-07-31   Initial release
1.1   2003-08-14   Added IOCTRL support, code-cleanup
1.1.1 2003-08-15   Fixed major problem (kernel oops) in the initialization of
                   the module
1.1.2 2003-08-15   Minor cleanup regarding speed when writing to the GPIOs
1.1.3 2003-08-18   Fixed problem with setting the direction of the GPIOs
                   (which was introduced with the previous speed improvement...)
1.1.4 2003-10-07   Fixed problem IOCTL calls causing segfault
1.2.0 2003-11-09   Rewrite to implement 4501 and 4801 driver with a common interface

*/

#ifndef __4501GPIO_H__
#define __4501GPIO_H__

#include "common.h"

/* 8 gpio pins  */
#define NUMBER_OF_PINS 8

#define OUR_NAME "4501gpio"

/* defines for IO-Addresses/offsets */
#define SC520_PIOPFS15_0        0xC20
#define SC520_PIOPFS31_16       0xC22
#define SC520_CSPFS             0xC24
#define SC520_CLKSEL            0xC26
#define SC520_DSCTL             0xC28
#define SC520_PIODIR15_0        0xC2A
#define SC520_PIODIR15_8        0xC2B
#define SC520_PIODIR31_16       0xC2C
#define SC520_PIODIR31_24       0xC2D
#define SC520_PIODATA15_0       0xC30
#define SC520_PIODATA15_9       0xC31
#define SC520_PIODATA31_16      0xC32
#define SC520_PIODATA31_24      0xC33
#define SC520_PIOSET15_0        0xC34
#define SC520_PIOSET15_8        0xC35
#define SC520_PIOSET31_16       0xC36
#define SC520_PIOSET31_24       0xC37
#define SC520_PIOCLR15_0        0xC38
#define SC520_PIOCLR15_8        0xC39
#define SC520_PIOCLR31_16       0xC3A
#define SC520_PIOCLR31_24       0xC3B
#define SC520_CBAR              0xFFFC

#define SC520_MMIO_SIZE         0xC3C

#define MMCR_BASE_DEFAULT 0xFFFEF000

static char* __our_name=OUR_NAME;
static int __number_of_pins=NUMBER_OF_PINS;


/*
Net45xx GPIO wiring

SC520 Pin   Function      | Pin Number | Function   SC520 Pin
                          +-----+------+
--          +3.3V  Power  |   1 | 2    | +5V Power  --
                          +-----+------+
PIO5, AD10  GPIO 0        |   3 | 4    | GPIO 1     PIO6, AE10
                          +-----+------+
PIO7, AF10  GPIO 2        |   5 | 6    | GPIO 3     PIO8, AF9
                          +-----+------+
PIO21, AF6  GPIO 4        |   7 | 8    ||GPIO 5     PIO22, AF5
                          +-----+------+
PIO11, AC9  GPIO 6        |   9 | 10   | GPIO 7     PIO12, AC8
                          +-----+------+
--          GND           |  11 | 12   | GND        --
                          +-----+------+
--          GND           |  13 | 14   | n-c        --
                          +-----+------+

*/

#endif
