/*
 **********************************************************************
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 **********************************************************************

Soekris Net48xx GPIO driver
Provides basic control (read and write) over the 12 general purpose
IO-pins of the Soekris Net4801 and the error-led (which happens to
be connected to another general purpose IO pin)
LEAF package by __PACKAGER__, __BUILDDATE__


Mapping between the
Accessing the GPIO pins (assuming all are output):
        echo 111111111111 > /proc/driver/soekris_gpio --> turns all ports on
        echo 000000000000 > /proc/driver/soekris_gpio --> turns all ports off
        echo 000000000100 > /proc/driver/soekris_gpio --> turns GPIO2 on,
                                                          everything else off

        cat /proc/driver/soekris_gpio   --> prints the current
                                            status of the gpio pins

        Note when GPIO are used as inputs:
        when "dangling" (not connected to anything)
        GPIO?? to GPIO?? are seen as "0"
        GPIO?? to GPIO?? are seen as "1"

        accessing the error LED:
        echo 1 > /proc/driver/soekris_error_led --> turns the led on

        cat /proc/driver/soekris_error_led
                                        --> prints the current status of the error led

Setting input or output for a pin
        echo 000010000001 /proc/driver/soekris_io_settings ->
                        turns GPIO7 and GPIO0 to output, all the others to input

        cat /proc/driver/soekris_io_settings -->prints the current io-settings

Devices:
        major ? (dynamic) minor 0 : set/read PIO0-7 (Net4801 mapping)
        major ? (dynamic) minor 1 : set/read PIO8-11 (Net4801 mapping)
        major ? (dynamic) minor 2 : set/read PIO0-11 (Net4801 mapping) (write low byte first)
        major ? (dynamic) minor 254 : set/read error_led (Net4801)
        Which major number has been assigned can be determined by looking at
        /proc/devices


Known IOCTL commands:
        PPCLAIM
        PPRELEASE
        PPFCONTROL
                    These don't do anything, but don't cause an error either

        PPDATADIR
                    Data line direction for GPIO0-7: non-zero for input mode.
                    This is _not_ "generic" at all, but very much hard-wired towards
                    being able to use an HD44780 LCD on the GPIO pins (in 4-bit mode)
                    and being able to do so using generic ppdev instructions

                    So, GPIO4-GPIO7 will _always_ be set to output after this call, only
                    GPIO0-GPIO3 are changed

                    This call does not change the input/output settings of GPIO8-11

        PPWDATA,PPRDATA
                    Reads/writes to the gpio0-7 (minor 0)
                    or the error led (minor 254)
                    (param is interpreted as a char)

        GPIORDDIRECTION,GPIOWRDIRECTION
                    "new" command (not from parport) - same effect
                    as using /proc/driver/soekris_io_settings
                    (param is interpreted as an int)

        GPIORDDATA,GPIOWRDATA
                    "new" command (not from parport)
                    read/write GPIO0-11 (param is interpreted as an int)


Credits/References
        Linux Device Drivers, 2nd Edition, O'Reilly, ISBN 0-59600-008-1
        http://www.tldp.org/LDP/lkmpg/
        soekris-tech mailing list
                (http://lists.soekris.com/mailman/listinfo/soekris-tech)
        Linux scx200 device driver
        Geode SC1100 Specification
        PC87366  Specification

(c) Copyright 2003   Martin Hejl <martin.hejl@guh-software.de>
                     G&H Softwareentwicklung GmbH

Changelog
1.2.0 2003-11-09   Added support for net4801

*/

#ifndef __4801GPIO_H__
#define __4801GPIO_H__

#include "common.h"

/* 12 gpio pins  */
#define NUMBER_OF_PINS 12

#define OUR_NAME "4801gpio"

#define SC1100_VENDOR_ID     0x100B
#define SC1100_F0_DEVICE_ID  0x0510

/* of set to use if we can't find the SIO at 0x2E */
#define SIO_BASE_OFFSET 0x20

/* defines for IO-Addresses/offsets */
#define SIO_GPIO_BANK0 sio_gpio_base
#define SIO_GPIO_BANK1 sio_gpio_base+0x4
#define SIO_GPIO_BANK2 sio_gpio_base+0x8

#define SIO_INDEX    (0x2E + offset_87336)
#define SIO_DATA     (0x2f + offset_87336)

#define SIO_DEV_SEL   0x7
#define SIO_DEV_ENB   0x30
#define SIO_GPIO_DEV  0x7
#define SIO_BASE_LADDR 0x61
#define SIO_BASE_HADDR 0x60
#define SIO_GPIO_PIN_SELECT    0xF0
#define SIO_GPIO_PIN_CONFIGURE 0xF1

#define CONFIG_GPIO_INPUT   0x04
#define CONFIG_GPIO_OUTPUT  0x03
#define SIO_SID       0x20    /* SuperI/O ID Register */


#define SIO_SID_VALUE 0xe9    /* Expected value in SuperI/O ID Register */

#define GEODE_GPIO_SIZE 0x2c  /* Size of GPIO register block */
#define GEODE_GPIO_DATA_IN 0x04
#define GEODE_GPIO_DATA_OUT 0x00

static unsigned short int sio_gpio_base=0;
static unsigned short int geode_gpio_base=0;
static unsigned short int offset_87336=0;


static char* __our_name=OUR_NAME;
static int __number_of_pins=NUMBER_OF_PINS;



/*

Pin assignment of the Net4801 GPIOs (connected to the 87226 GPIO pins
and _NOT_ to the GPIO pins of the GEODE processor)

                     +----+----+
               3,3 V |  1 |  2 | 5 V
                     +----+----+
GPIO0   Port 2 Bit 0 |  3 |  4 | Port 2 Bit 1   GPIO1
                     +----+----+
GPIO2   Port 2 Bit 2 |  5 |  6 | Port 2 Bit 3   GPIO3
                     +----+----+
GPIO4   Port 2 Bit 4 |  7 |  8 | Port 2 Bit 5   GPIO5
                     +----+----+
GPIO6   Port 2 Bit 6 |  9 | 10 | Port 2 Bit 7   GPIO7
                     +----+----+
                 GND | 11 | 12 | Port 0 Bit 4   GPIO8
                     +----+----+
GPIO9   Port 0 Bit 5 | 13 | 14 | GND
                     +----+----+
GPIO10  Port 1 Bit 3 | 15 | 16 | Port 1 Bit 2   GPIO11
                     +----+----+
                 GND | 17 | 18 | unknown (docs say GND, but is 3.3V on my box)
                     +----+----+
               3,3 V | 19 | 20 | GND
                     +----+----+

Error_LED => GEODE GPIO20

*/

#endif
