/*
 Soekris Netxxxx GPIO driver
 Common functions shared between the net4501 and net4801 driver

 **********************************************************************
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 **********************************************************************
 *
 */


#include "common.h"

extern struct gpio_operations driver_ops;

extern int __number_of_pins;
extern char* __our_name;


//------------------------------------------------
static ssize_t net4xxx_gpio_write(struct file *file, const char *data,
                                 size_t count, loff_t *ppos)
{
        unsigned m = minor(file->f_dentry->d_inode->i_rdev);
        size_t i;
        unsigned int value=0;
        char port_status=0;
        ssize_t ret = 0;

        if (ppos != &file->f_pos)
                return -ESPIPE;

        if (down_interruptible(&gpio_sema))
                return -ERESTARTSYS;

        for (i = 0; i < count; ++i) {
                if(copy_from_user(&port_status, data+i, sizeof(char))) {
                        ret = -EFAULT;
                        goto out;
                }

                switch(m) {
                        case MINOR_BYTE:
                                driver_ops.write8Bit(port_status);
                                break;

                        case MINOR_FULL:
                                if ((i&1) == 0) {
                                        value = port_status;
                                } else {
                                        value |= ((port_status&0x0F) << 8);
                                        driver_ops.write16Bit(value);
                                }
                                break;

                        case MINOR_LED:
                                driver_ops.writeErrorLed(port_status);
                                break;

                        default:
                                ret=-EINVAL;
                                goto out;
                }
                data++;
                *ppos = *ppos+1;
        }
        ret = count;

out:
        up(&gpio_sema);
        return ret;
}

static ssize_t net4xxx_gpio_read(struct file *file, char *buf,
                                size_t count, loff_t *ppos)
{

        unsigned m = minor(file->f_dentry->d_inode->i_rdev);
        unsigned long value=0;
        size_t bytes_read=0;
        ssize_t ret = 0;
        char port_status=0;

        if (count == 0)
                return bytes_read;

        if (ppos != &file->f_pos)
                return -ESPIPE;

        if (down_interruptible(&gpio_sema))
                return -ERESTARTSYS;

        switch(m) {
                case MINOR_BYTE:
                        value = driver_ops.read8Bit();
                        port_status = (char)(value & 0xFF);
                        if (copy_to_user(buf, &port_status, sizeof(char))) {
                                ret = -EFAULT;
                                goto out;
                        }
                        bytes_read++;
                        buf++;
                        *ppos = *ppos+1;
                        ret = bytes_read;
                        break;

                case MINOR_FULL:
                        value = driver_ops.read16Bit();
                        port_status = (char)(value & 0xFF);
                        if (copy_to_user(buf, &port_status, sizeof(char))) {
                                ret = -EFAULT;
                                goto out;
                        }
                        bytes_read++;
                        buf++;
                        *ppos = *ppos+1;

                        if (count>1) {
                                port_status = (char)((value>>8) & 0xF);
                                if (copy_to_user(buf, &port_status, sizeof(char))) {
                                        ret = -EFAULT;
                                        goto out;
                                }
                                bytes_read++;
                                buf++;
                                *ppos = *ppos+1;
                        }

                        ret = bytes_read;
                        break;

                case MINOR_LED:
                        port_status = driver_ops.readErrorLed();
                        if (copy_to_user(buf, &port_status, sizeof(char))) {
                                ret = -EFAULT;
                                goto out;
                        }
                        bytes_read++;
                        buf++;
                        *ppos = *ppos+1;
                        ret = bytes_read;
                        break;

                default:
                        ret = -EFAULT;
                        goto out;
        }

out:
        up(&gpio_sema);
        return ret;

}


static int net4xxx_gpio_open(struct inode *inode, struct file *file)
{
        unsigned m = minor(inode->i_rdev);
        if (m!=MINOR_BYTE && m!=MINOR_FULL && m!=MINOR_LED )
                return -EINVAL;

        return 0;
}

static int net4xxx_gpio_release(struct inode *inode, struct file *file)
{
        return 0;
}

//------------------------------------------------

// PROC file
int procfile_gpio_read(
                __attribute__ ((unused)) char *buffer,
                char **buffer_location,
                off_t offset,
                __attribute__ ((unused)) int buffer_length,
                __attribute__ ((unused)) int *eof,
                __attribute__ ((unused)) void *data)

{
        int len; /* The number of bytes actually used */
        /* This is static so it will still be in memory
        * when we leave this function */
        static char output_buffer[MAX_NUMBER_OF_PINS+2];
        unsigned int port_status=0;

        /* We give all of our information in one go, so if the
        * user asks us if we have more information the
        * answer should always be no.
        *
        * This is important because the standard read
        * function from the library would continue to issue
        * the read system call until the kernel replies
        * that it has no more information, or until its
        * buffer is filled.
        */
        if (offset > 0)
        {
                return 0;
        }

        MOD_INC_USE_COUNT;

        /* Get the status of the gpio ports */
        if (driver_ops.read16Bit == NULL) {
                port_status = driver_ops.read8Bit();
        } else {
                port_status = driver_ops.read16Bit();
        }

        driver_ops.toProcString(port_status,output_buffer,__number_of_pins);

        /* Tell the function which called us where the
        * buffer is */
        *buffer_location = output_buffer;

        len = __number_of_pins+1;

        /* Return the length */
        MOD_DEC_USE_COUNT;
        return len;

}

int procfile_gpio_write(
        __attribute__ ((unused)) struct file *file,
        const char *buf,
        unsigned long count,
        __attribute__ ((unused)) void *data)
{
        static int len;
        static char new_gpio_state[MAX_NUMBER_OF_PINS+1];
        static unsigned int  gpio_state;


        if (count==0) return 0;
        MOD_INC_USE_COUNT;

        if(count > __number_of_pins)
        {
                len = __number_of_pins;
        } else {
                len = count;
        }

        if(copy_from_user(new_gpio_state, buf, len)) {
                MOD_DEC_USE_COUNT;
                return -EFAULT;
        }

        gpio_state = driver_ops.fromProcString(new_gpio_state,count);

        if (driver_ops.write16Bit==NULL) {
                driver_ops.write8Bit(gpio_state);
        } else {
                driver_ops.write16Bit(gpio_state);
        }

        MOD_DEC_USE_COUNT;

        return len;

}

int procfile_settings_read(
                __attribute__ ((unused)) char *buffer,
                char **buffer_location,
                off_t offset,
                __attribute__ ((unused)) int buffer_length,
                __attribute__ ((unused)) int *eof,
                __attribute__ ((unused)) void *data)

{
        int len; /* The number of bytes actually used */
        static char output_buffer[MAX_NUMBER_OF_PINS+2];
        unsigned int port_status=0;

        if (offset > 0)
                return 0;

        MOD_INC_USE_COUNT;
        if (driver_ops.set16BitDirection == NULL) {
                port_status = driver_ops.get8BitDirection();
        } else {
                port_status = driver_ops.get16BitDirection();
        }

        driver_ops.toProcString(port_status,output_buffer,__number_of_pins);

        len = __number_of_pins+1;

        *buffer_location = output_buffer;
        MOD_DEC_USE_COUNT;
        return len;

}

int procfile_settings_write(
        __attribute__ ((unused)) struct file *file,
        const char *buf,
        unsigned long count,
        __attribute__ ((unused)) void *data)
{
        static int len;
        static char new_gpio_state[MAX_NUMBER_OF_PINS+1];
        static unsigned int  gpio_state;


        if (count==0) return 0;
        MOD_INC_USE_COUNT;

        if(count > __number_of_pins)
        {
                len = __number_of_pins;
        } else {
                len = count;
        }

        if(copy_from_user(new_gpio_state, buf, len)) {
                MOD_DEC_USE_COUNT;
                return -EFAULT;
        }


        gpio_state = driver_ops.fromProcString(new_gpio_state,len);

        if (driver_ops.set16BitDirection==NULL) {
                driver_ops.set8BitDirection(gpio_state);
        } else {
                driver_ops.set16BitDirection(gpio_state);
        }

        MOD_DEC_USE_COUNT;

        return len;
}

int procfile_led_read(
                __attribute__ ((unused)) char *buffer,
                char **buffer_location,
                off_t offset,
                __attribute__ ((unused)) int buffer_length,
                __attribute__ ((unused)) int *eof,
                __attribute__ ((unused)) void *data)

{
        static char output_buffer[3];

        if (offset > 0)
        {

                return 0;
        }

        MOD_INC_USE_COUNT;

        if (driver_ops.readErrorLed() == 0) {
                output_buffer[0] = '0';
        } else {
                output_buffer[0] = '1';
        }

        output_buffer[1] = '\n';
        output_buffer[2] = 0;

        *buffer_location = output_buffer;
        MOD_DEC_USE_COUNT;
        return 2;

}

int procfile_led_write(
        __attribute__ ((unused)) struct file *file,
        const char *buf,
        unsigned long count,
        __attribute__ ((unused)) void *data)
{
        static int len;
        static char new_led_state[MAX_NUMBER_OF_PINS+1];

        if (count==0) return 0;

        MOD_INC_USE_COUNT;

        if(count > __number_of_pins)
        {
                len = __number_of_pins;
        } else {
                len = count;
        }

        if(copy_from_user(new_led_state, buf, len)) {
                MOD_DEC_USE_COUNT;
                return -EFAULT;
        }

        if (new_led_state[0] == '1') {
                driver_ops.writeErrorLed(1);
        } else {
                driver_ops.writeErrorLed(0);
        }


        MOD_DEC_USE_COUNT;

        return len;

}
//------------------------------------------------


int net4xxx_gpio_ioctl(struct inode *inode, struct file *filp,
                unsigned int cmd, unsigned long arg)
{
        int err = 0;
        int ret = 0;
        int value=0;

        /*
        * extract the type and number bitfields, and don't decode
        * wrong cmds: return ENOTTY (inappropriate ioctl) before access_ok()
        */
        if (_IOC_TYPE(cmd) != PP_IOCTL) return -ENOTTY;

        /*
        * the direction is a bitmask, and VERIFY_WRITE catches R/W
        * transfers. `Type' is user-oriented, while
        * access_ok is kernel-oriented, so the concept of "read" and
        * "write" is reversed
        */
        if (_IOC_DIR(cmd) & _IOC_READ)
                err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
        else if (_IOC_DIR(cmd) & _IOC_WRITE)
                err =  !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));

        if (err) return -EFAULT;

        MOD_INC_USE_COUNT;

        switch(MINOR(inode->i_rdev)) {
        case MINOR_BYTE:        // /dev/gpio0 controls the 8 gpio ports
        case MINOR_FULL:        // /dev/gpio1 controls the 12 gpio ports
                switch(cmd) {
                case PPCLAIM:
                case PPRELEASE:
                case PPFCONTROL:
                        break;

                case PPWDATA:
                        ret = __get_user(value, (unsigned char *)arg);

                        if (ret==0) {
                                driver_ops.write8Bit(value);
                        }
                        break;

                case PPRDATA:
                        value = driver_ops.read8Bit();
                        ret = __put_user(value, (unsigned char *)arg);
                        break;

                case PPDATADIR:
                        ret = __get_user(value, (unsigned char *)arg);

                        /* linux/ppdev.h define PPDATADIR as
                           "Data line direction: non-zero for input mode."

                           For gpios, the logic is reversed - bit=1 == output

                           This is _not_ "generic" at all, but very much hard-wired towards
                           being able to use an HD44780 LCD on the GPIO pins (in 4-bit mode)
                           and being able to do so using generic ppdev instructions

                           So, GPIO4-GPIO7 will _always_ be set to output for this call, only
                           GPIO0-GPIO3 are changed

                         */

                        if (ret==0) {
                                if (value==0) {
                                        driver_ops.set8BitDirection(0xFF);
                                } else {
                                        driver_ops.set8BitDirection(0xF0);
                                }
                        } else {
                                printk("%s: ret=%x\n", __our_name, ret);
                        }


                        break;

                case GPIORDDIRECTION:
                        if (MINOR(inode->i_rdev)==MINOR_BYTE) {
                                value = driver_ops.get8BitDirection();
                        } else {
                                value = driver_ops.get16BitDirection();
                        }
                        ret = __put_user(value, (unsigned int *)arg);
                        break;

                case GPIOWRDIRECTION:
                        ret = __get_user(value, (unsigned int *)arg);

                        if (ret==0) {
                                if (MINOR(inode->i_rdev)==MINOR_BYTE) {
                                        driver_ops.set8BitDirection(value);
                                } else {
                                        driver_ops.set16BitDirection(value);
                                }
                        }

                        break;

                case GPIORDDATA:
                        if (MINOR(inode->i_rdev)==MINOR_BYTE) {
                                value = driver_ops.read8Bit();
                        } else {
                                value = driver_ops.read16Bit();
                        }

                        ret = __put_user(value, (unsigned int *)arg);
                        break;

                case GPIOWRDATA:
                        ret = __get_user(value, (unsigned int *)arg);

                        if (ret==0) {
                                if (MINOR(inode->i_rdev)==MINOR_BYTE) {
                                        driver_ops.write8Bit(value);
                                } else {
                                        driver_ops.write16Bit(value);
                                }                        }
                        break;

                default:
                        return -ENOTTY;
                }
                break;

        case MINOR_LED:        // /dev/gpio254 controls the error led
                switch(cmd) {
                case PPCLAIM:
                case PPRELEASE:
                case PPFCONTROL:
                case PPDATADIR: /* Doesn't work the same way */
                        break;

                case PPWDATA:
                        ret = __get_user(value, (unsigned char *)arg);

                        if (ret==0) {
                                driver_ops.writeErrorLed(value);
                        }

                        break;

                case PPRDATA:
                        value = driver_ops.readErrorLed();
                        ret = __put_user(value, (unsigned char *)arg);
                        break;

                default:
                        return -ENOTTY;
                }
                break;
        }

        MOD_DEC_USE_COUNT;

        return ret;

}
static struct file_operations gpio_dispatch_fops = {
        .owner   = THIS_MODULE,
        .write   = net4xxx_gpio_write,
        .read    = net4xxx_gpio_read,
        .open    = net4xxx_gpio_open,
        .release = net4xxx_gpio_release,
        .ioctl   = net4xxx_gpio_ioctl,
};


static int __init common_init(void)
{
        int result;
        struct proc_dir_entry *GPIO_Proc_File;
        struct proc_dir_entry *ErrorLED_Proc_File;
        struct proc_dir_entry *Settings_Proc_File;

        result = driver_ops.init();
        if (result!=0) return result;

        result = register_chrdev(gpio_major, __our_name, &gpio_dispatch_fops);
        if (result < 0) {
                printk(KERN_WARNING "%s: can't get major %d\n",__our_name, gpio_major);
                return result;
        }

        if (gpio_major == 0) gpio_major = result; /* dynamic */

        GPIO_Proc_File = create_proc_entry(
                GPIO_PROC_FILENAME,
                0, /*S_IFREG | S_IRUGO,*/
                NULL);

        if(GPIO_Proc_File == NULL) {
                printk(KERN_ERR "%s: Could not register " GPIO_PROC_FILENAME  ". Terminating\n", __our_name);
                unregister_chrdev(gpio_major, __our_name);
                driver_ops.cleanup();
                return -ENOMEM;
        } else {
                GPIO_Proc_File->read_proc=procfile_gpio_read;
                GPIO_Proc_File->write_proc=procfile_gpio_write;
        }

        ErrorLED_Proc_File = create_proc_entry(
                LED_PROC_FILENAME,
                0,
                NULL);

        if(ErrorLED_Proc_File == NULL) {
                printk(KERN_ERR "%s:Could not register " LED_PROC_FILENAME ". Terminating\n", __our_name);
                remove_proc_entry(GPIO_PROC_FILENAME, NULL);
                unregister_chrdev(gpio_major, __our_name);
                driver_ops.cleanup();
                return -ENOMEM;
        } else {
                ErrorLED_Proc_File->read_proc=procfile_led_read;
                ErrorLED_Proc_File->write_proc=procfile_led_write;
        }

        Settings_Proc_File = create_proc_entry(
                SETTINGS_PROC_FILENAME,
                0,
                NULL);

        if(Settings_Proc_File == NULL) {
                printk(KERN_ERR "%s: Could not register " SETTINGS_PROC_FILENAME ". Terminating\n", __our_name);
                remove_proc_entry(GPIO_PROC_FILENAME, NULL);
                remove_proc_entry(LED_PROC_FILENAME, NULL);
                unregister_chrdev(gpio_major, __our_name);
                driver_ops.cleanup();
                return -ENOMEM;
        } else {
                Settings_Proc_File->read_proc=procfile_settings_read;
                Settings_Proc_File->write_proc=procfile_settings_write;
        }

        SET_MODULE_OWNER(&gpio_dispatch_fops);

        sema_init(&gpio_sema, 1);

        /* set gpio0-7 to output */
        driver_ops.set8BitDirection(0xFF);

        return(0);
}

void __exit common_cleanup(void)
{
        unregister_chrdev(gpio_major, __our_name);
        remove_proc_entry(GPIO_PROC_FILENAME, NULL);
        remove_proc_entry(LED_PROC_FILENAME, NULL);
        remove_proc_entry(SETTINGS_PROC_FILENAME, NULL);
        driver_ops.cleanup();

}

//------------------------------------------------

module_init(common_init);
module_exit(common_cleanup);

MODULE_AUTHOR("Martin Hejl");
MODULE_DESCRIPTION("Soekris net4xxx GPIO / Error LED driver");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
